@@ -13,14 +13,12 @@
 # limitations under the License.
 
 import functools
-import importlib.util
 import io
 import json
 import logging
 import os.path
 import pathlib
 import sys
-import types
 
 import cloudevents.exceptions as cloud_exceptions
 import flask
@@ -28,20 +26,15 @@ import werkzeug
 
 from cloudevents.http import from_http, is_binary
 
-from functions_framework import event_conversion
+from functions_framework import _function_registry, event_conversion
 from functions_framework.background_event import BackgroundEvent
 from functions_framework.exceptions import (
     EventConversionException,
     FunctionsFrameworkException,
-    InvalidConfigurationException,
-    InvalidTargetTypeException,
     MissingSourceException,
-    MissingTargetException,
 )
 from google.cloud.functions.context import Context
 
-DEFAULT_SOURCE = os.path.realpath("./main.py")
-DEFAULT_SIGNATURE_TYPE = "http"
 MAX_CONTENT_LENGTH = 10 * 1024 * 1024
 
 _FUNCTION_STATUS_HEADER_FIELD = "X-Google-Status"
@@ -63,6 +56,32 @@ class _LoggingHandler(io.TextIOWrapper):
         return self.stderr.write(json.dumps(payload) + "\n")
 
 
+def cloud_event(func):
+    """Decorator that registers cloudevent as user function signature type."""
+    _function_registry.REGISTRY_MAP[
+        func.__name__
+    ] = _function_registry.CLOUDEVENT_SIGNATURE_TYPE
+
+    @functools.wraps(func)
+    def wrapper(*args, **kwargs):
+        return func(*args, **kwargs)
+
+    return wrapper
+
+
+def http(func):
+    """Decorator that registers http as user function signature type."""
+    _function_registry.REGISTRY_MAP[
+        func.__name__
+    ] = _function_registry.HTTP_SIGNATURE_TYPE
+
+    @functools.wraps(func)
+    def wrapper(*args, **kwargs):
+        return func(*args, **kwargs)
+
+    return wrapper
+
+
 def setup_logging():
     logging.getLogger().setLevel(logging.INFO)
     info_handler = logging.StreamHandler(sys.stdout)
@@ -82,13 +101,13 @@ def _http_view_func_wrapper(function, request):
     return view_func
 
 
-def _run_cloudevent(function, request):
+def _run_cloud_event(function, request):
     data = request.get_data()
     event = from_http(request.headers, data)
     function(event)
 
 
-def _cloudevent_view_func_wrapper(function, request):
+def _cloud_event_view_func_wrapper(function, request):
     def view_func(path):
         ce_exception = None
         event = None
@@ -106,7 +125,7 @@ def _cloudevent_view_func_wrapper(function, request):
 
         # Not a CloudEvent. Try converting to a CloudEvent.
         try:
-            function(event_conversion.background_event_to_cloudevent(request))
+            function(event_conversion.background_event_to_cloud_event(request))
         except EventConversionException as e:
             flask.abort(
                 400,
@@ -125,9 +144,9 @@ def _cloudevent_view_func_wrapper(function, request):
 
 def _event_view_func_wrapper(function, request):
     def view_func(path):
-        if event_conversion.is_convertable_cloudevent(request):
+        if event_conversion.is_convertable_cloud_event(request):
             # Convert this CloudEvent to the equivalent background event data and context.
-            data, context = event_conversion.cloudevent_to_background_event(request)
+            data, context = event_conversion.cloud_event_to_background_event(request)
             function(data, context)
         elif is_binary(request.headers):
             # Support CloudEvents in binary content mode, with data being the
@@ -156,6 +175,56 @@ def _event_view_func_wrapper(function, request):
     return view_func
 
 
+def _configure_app(app, function, signature_type):
+    # Mount the function at the root. Support GCF's default path behavior
+    # Modify the url_map and view_functions directly here instead of using
+    # add_url_rule in order to create endpoints that route all methods
+    if signature_type == _function_registry.HTTP_SIGNATURE_TYPE:
+        app.url_map.add(
+            werkzeug.routing.Rule("/", defaults={"path": ""}, endpoint="run")
+        )
+        app.url_map.add(werkzeug.routing.Rule("/robots.txt", endpoint="error"))
+        app.url_map.add(werkzeug.routing.Rule("/favicon.ico", endpoint="error"))
+        app.url_map.add(werkzeug.routing.Rule("/<path:path>", endpoint="run"))
+        app.view_functions["run"] = _http_view_func_wrapper(function, flask.request)
+        app.view_functions["error"] = lambda: flask.abort(404, description="Not Found")
+        app.after_request(read_request)
+    elif signature_type == _function_registry.BACKGROUNDEVENT_SIGNATURE_TYPE:
+        app.url_map.add(
+            werkzeug.routing.Rule(
+                "/", defaults={"path": ""}, endpoint="run", methods=["POST"]
+            )
+        )
+        app.url_map.add(
+            werkzeug.routing.Rule("/<path:path>", endpoint="run", methods=["POST"])
+        )
+        app.view_functions["run"] = _event_view_func_wrapper(function, flask.request)
+        # Add a dummy endpoint for GET /
+        app.url_map.add(werkzeug.routing.Rule("/", endpoint="get", methods=["GET"]))
+        app.view_functions["get"] = lambda: ""
+    elif signature_type == _function_registry.CLOUDEVENT_SIGNATURE_TYPE:
+        app.url_map.add(
+            werkzeug.routing.Rule(
+                "/", defaults={"path": ""}, endpoint=signature_type, methods=["POST"]
+            )
+        )
+        app.url_map.add(
+            werkzeug.routing.Rule(
+                "/<path:path>", endpoint=signature_type, methods=["POST"]
+            )
+        )
+
+        app.view_functions[signature_type] = _cloud_event_view_func_wrapper(
+            function, flask.request
+        )
+    else:
+        raise FunctionsFrameworkException(
+            "Invalid signature type: {signature_type}".format(
+                signature_type=signature_type
+            )
+        )
+
+
 def read_request(response):
     """
     Force the framework to read the entire request before responding, to avoid
@@ -174,21 +243,8 @@ def crash_handler(e):
 
 
 def create_app(target=None, source=None, signature_type=None):
-    # Get the configured function target
-    target = target or os.environ.get("FUNCTION_TARGET", "")
-    # Set the environment variable if it wasn't already
-    os.environ["FUNCTION_TARGET"] = target
-
-    if not target:
-        raise InvalidConfigurationException(
-            "Target is not specified (FUNCTION_TARGET environment variable not set)"
-        )
-
-    # Get the configured function source
-    source = source or os.environ.get("FUNCTION_SOURCE", DEFAULT_SOURCE)
-
-    # Python 3.5: os.path.exist does not support PosixPath
-    source = str(source)
+    target = _function_registry.get_function_target(target)
+    source = _function_registry.get_function_source(source)
 
     # Set the template folder relative to the source path
     # Python 3.5: join does not support PosixPath
@@ -201,126 +257,43 @@ def create_app(target=None, source=None, signature_type=None):
             )
         )
 
-    # Get the configured function signature type
-    signature_type = signature_type or os.environ.get(
-        "FUNCTION_SIGNATURE_TYPE", DEFAULT_SIGNATURE_TYPE
-    )
-    # Set the environment variable if it wasn't already
-    os.environ["FUNCTION_SIGNATURE_TYPE"] = signature_type
+    source_module, spec = _function_registry.load_function_module(source)
 
-    # Load the source file:
-    # 1. Extract the module name from the source path
-    realpath = os.path.realpath(source)
-    directory, filename = os.path.split(realpath)
-    name, extension = os.path.splitext(filename)
-
-    # 2. Create a new module
-    spec = importlib.util.spec_from_file_location(name, realpath)
-    source_module = importlib.util.module_from_spec(spec)
-
-    # 3. Add the directory of the source to sys.path to allow the function to
-    # load modules relative to its location
-    sys.path.append(directory)
-
-    # 4. Add the module to sys.modules
-    sys.modules[name] = source_module
-
-    # 5. Create the application
-    app = flask.Flask(target, template_folder=template_folder)
-    app.config["MAX_CONTENT_LENGTH"] = MAX_CONTENT_LENGTH
-    app.register_error_handler(500, crash_handler)
+    # Create the application
+    _app = flask.Flask(target, template_folder=template_folder)
+    _app.config["MAX_CONTENT_LENGTH"] = MAX_CONTENT_LENGTH
+    _app.register_error_handler(500, crash_handler)
     global errorhandler
-    errorhandler = app.errorhandler
+    errorhandler = _app.errorhandler
 
-    # 6. Handle legacy GCF Python 3.7 behavior
+    # Handle legacy GCF Python 3.7 behavior
     if os.environ.get("ENTRY_POINT"):
-        os.environ["FUNCTION_TRIGGER_TYPE"] = signature_type
         os.environ["FUNCTION_NAME"] = os.environ.get("K_SERVICE", target)
-        app.make_response_original = app.make_response
+        _app.make_response_original = _app.make_response
 
         def handle_none(rv):
             if rv is None:
                 rv = "OK"
-            return app.make_response_original(rv)
+            return _app.make_response_original(rv)
 
-        app.make_response = handle_none
+        _app.make_response = handle_none
 
         # Handle log severity backwards compatibility
         sys.stdout = _LoggingHandler("INFO", sys.stderr)
         sys.stderr = _LoggingHandler("ERROR", sys.stderr)
         setup_logging()
 
-    # 7. Execute the module, within the application context
-    with app.app_context():
+    # Execute the module, within the application context
+    with _app.app_context():
         spec.loader.exec_module(source_module)
 
-    # Extract the target function from the source file
-    if not hasattr(source_module, target):
-        raise MissingTargetException(
-            "File {source} is expected to contain a function named {target}".format(
-                source=source, target=target
-            )
-        )
-    function = getattr(source_module, target)
+    # Get the configured function signature type
+    signature_type = _function_registry.get_func_signature_type(target, signature_type)
+    function = _function_registry.get_user_function(source, source_module, target)
 
-    # Check that it is a function
-    if not isinstance(function, types.FunctionType):
-        raise InvalidTargetTypeException(
-            "The function defined in file {source} as {target} needs to be of "
-            "type function. Got: invalid type {target_type}".format(
-                source=source, target=target, target_type=type(function)
-            )
-        )
+    _configure_app(_app, function, signature_type)
 
-    # Mount the function at the root. Support GCF's default path behavior
-    # Modify the url_map and view_functions directly here instead of using
-    # add_url_rule in order to create endpoints that route all methods
-    if signature_type == "http":
-        app.url_map.add(
-            werkzeug.routing.Rule("/", defaults={"path": ""}, endpoint="run")
-        )
-        app.url_map.add(werkzeug.routing.Rule("/robots.txt", endpoint="error"))
-        app.url_map.add(werkzeug.routing.Rule("/favicon.ico", endpoint="error"))
-        app.url_map.add(werkzeug.routing.Rule("/<path:path>", endpoint="run"))
-        app.view_functions["run"] = _http_view_func_wrapper(function, flask.request)
-        app.view_functions["error"] = lambda: flask.abort(404, description="Not Found")
-        app.after_request(read_request)
-    elif signature_type == "event":
-        app.url_map.add(
-            werkzeug.routing.Rule(
-                "/", defaults={"path": ""}, endpoint="run", methods=["POST"]
-            )
-        )
-        app.url_map.add(
-            werkzeug.routing.Rule("/<path:path>", endpoint="run", methods=["POST"])
-        )
-        app.view_functions["run"] = _event_view_func_wrapper(function, flask.request)
-        # Add a dummy endpoint for GET /
-        app.url_map.add(werkzeug.routing.Rule("/", endpoint="get", methods=["GET"]))
-        app.view_functions["get"] = lambda: ""
-    elif signature_type == "cloudevent":
-        app.url_map.add(
-            werkzeug.routing.Rule(
-                "/", defaults={"path": ""}, endpoint=signature_type, methods=["POST"]
-            )
-        )
-        app.url_map.add(
-            werkzeug.routing.Rule(
-                "/<path:path>", endpoint=signature_type, methods=["POST"]
-            )
-        )
-
-        app.view_functions[signature_type] = _cloudevent_view_func_wrapper(
-            function, flask.request
-        )
-    else:
-        raise FunctionsFrameworkException(
-            "Invalid signature type: {signature_type}".format(
-                signature_type=signature_type
-            )
-        )
-
-    return app
+    return _app
 
 
 class LazyWSGIApp:
